
#include <stdlib.h>
#include <math.h>
#include <complex.h>

typedef struct same_s {
	double complex z;
	double complex fz;
	struct same_s *next;
} same_t;

typedef struct it_s {
	double complex z;
	double complex fz;
	unsigned int size;
	struct it_s *next;
	struct same_s *same;
} it_t;

typedef struct list_s {
	double complex z;
	double complex fz;
	unsigned int size;
	it_t *head;
	double eps;
} list_t;

static list_t list;

void lavg_init(double eps) {
	list.z = NULL;
	list.fz = NULL;
	list.head = NULL;
	list.size = 0;
	list.eps = eps;
}

void lavg_destroy() {
	it_t *i = list.head, *in;
	same_t *s, *sn;
	
	while (i != NULL) {
		in = i->next;
		s = i->same;
		while (s != NULL) {
			sn = s->next;
			free(s);
			s = sn;
		}
		free(i);
		i = in;
	}
}

void lavg_insert(double complex z, double complex fz) {
	it_t *i;
	same_t *s, *ns;
	double dfz;
	
	ns = malloc(sizeof(same_t));
	ns->z = z;
	ns->fz = fz;
	ns->next = NULL;
	
	for (i=list.head; i != NULL; i = i->next)
		if ((dfz = cabs(fz - i->fz)) <= list.eps)
			break;
	
	if (i == NULL) {
		i = malloc(sizeof(it_t));
		i->same = ns;
		i->z = z;
		i->fz = fz;
		i->size = 1;
		list.size++;
		i->next = list.head;
		list.head = i;
	} else {
		i->size++;
		s->next = i->same;
		i->same = ns;
		// todo: durchschnitt -> i, gewicht: 1/cabs(fz)
	}
}
