
#define _XOPEN_SOURCE	/* for M_PI, we could also take acos(-1), but well... */

#include <math.h>
#include <complex.h>
#include <stdbool.h>

#include "cgamma.h"

/* Lanczos approximation for the complex plane calculated with vpa digits(128) */
static const double c[] = {
		     1.000000000000000174663,
		  5716.400188274341379136,
		-14815.30426768413909044,
		 14291.49277657478554025,
		 -6348.160217641458813289,
		  1301.608286058321874105,
		  -108.1767053514369634679,
		     2.605696505611755827729,
		    -0.7423452510201416151527e-2,
		     0.5384136432509564062961e-7,
		    -0.4023533141268236372067e-8 };

double complex lcgamma(double complex z) {
	int k;
	static double twopi = M_PI + M_PI;
	double zre = creal(z);
	bool neg = zre < 0.0;
	
	if (neg)
		z = -z;
	
	/* Num Recipes used g=9 for a lower order approximation */
	int g = 9;
	
	double complex t = z + g;
	double complex s = 0;
	for (k = g + 1; k > 0; k--) {
		s += c[k] / t;
		t -= 1;
	}
	s += c[0];
	double complex ss = z + g - 0.5;
	return clog(s * sqrt(twopi)) + (z - 0.5) * clog(ss) - ss;
}

// from http://home.att.net/~numericana/answer/info/godfrey.htm
double complex cgamma(double complex z, double complex *log_of_gamma) {
	double complex s, f;
	double zre;
	bool neg;
	
	s = lcgamma(z);
	
	if (log_of_gamma) *log_of_gamma = s;
	
	f = cexp(s);
	zre = creal(z);
	neg = zre < 0.0;
	if (neg)
		f = -M_PI / (z * f * csin(M_PI * z));
	
	// make results exact for integer arguments
	if (cimag(z) == 0.0 && rint(zre) == zre) {
		if (neg || zre == 0.0) {
			f = HUGE_VAL;
		} else {
			int pp = 1;
			int zint;
			for (zint = 1; zint <= zre; zint++) {
				if (zre == zint)
					f = pp;
				pp *= zint;
			}
		}
	}
	
	return f;
}
