
#define _XOPEN_SOURCE 500

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <complex.h>

#include "formparse.h"
#include "cgamma.h"

#define NEXTERN
#include "eval.h"
#undef NEXTERN

/* todo: instead of an explicit stack, generate the code from the program's .text segment and put it together as
 *       stack-"elements", so that the checks for (s < end), the switch (s->op), s++ and maybe setting prev can be saved */
#define _prefix __m			/* prefix for all the labels used to identify the different calculation sections */
#define str(x) #x			/* stringify the argument, meaning: put it into quotation marks */
#define xstr(x) str(x)		/* need indirection to expand the argument if it is a macro */
#define epair(name) _prefix ## name ## 0, _prefix ## name ## 1
extern char
	epair(var), epair(num), epair(numf), epair(fun),
	epair(fac), epair(exp), epair(mul), epair(div),
	epair(add), epair(sub);
#undef epair
#define rpair(name) { &_prefix ## name ## 0, &_prefix ## name ## 1 }
static char *marks[][2] = {
	rpair(var), rpair(num), rpair(numf), rpair(fun),
	rpair(fac), rpair(exp), rpair(mul), rpair(div),
	rpair(add), rpair(sub)
};
#undef rpair

static double complex ___cpow(double complex a, double complex z) {
	double complex w;
	double x, y, r, theta, absa, arga;
	
	x = creal(z);
	y = cimag(z);
	absa = cabs(a);
	if (absa == 0.0)
		return (0.0 + 0.0 * I);
	
	arga = carg(a);
	r = pow(absa, x);
	theta = x * arga;
	if (y != 0.0) {
		r = r * exp(-y * arga);
		theta = theta + y * log(absa);
	}
	w = r * cos(theta) + (r * sin(theta)) * I;
	
	return w;
}

#define amark0(name) do { __asm__ volatile(xstr(_prefix) #name "0:")
#define amark1(name) __asm__ volatile(xstr(_prefix) #name "1:"); } while(0)

typedef stackit_t * __fastcall (*_c_func)(stackit_t * const);
#define _calcop(abbr) stackit_t * __fastcall _calc_ ## abbr(stackit_t * const s)

_calcop(var) {
	s->data.val = (*s->var)[s->data.idx];
	return s;
}

_calcop(num) {
	return s;
}
_calcop(numf) {
	return s;
}

_calcop(fun) {
	stackit_t *lo = s->prev;
	lo->data.val = s->data.fun.c(lo->data.val);
	return lo;
}

_calcop(fac) {
	stackit_t *lo = s->prev;
	lo->data.val = cgamma(lo->data.val, NULL);
	return lo;
}

_calcop(exp) {
	stackit_t *t = s->prev;
	stackit_t *lo = t->prev;
	lo->data.val = ___cpow(lo->data.val, t->data.val);
	return lo;
}

#define _calcelem(op, abbr) _calcop(abbr) { \
		stackit_t *t = s->prev; \
		stackit_t *lo = t->prev; \
		lo->data.val op ## = t->data.val; \
		return lo; \
	}
_calcelem(*, mul)
_calcelem(/, div)
_calcelem(+, add)
_calcelem(-, sub)
#undef _calcelem

inline static void __fastcall _calc_interprete2(stackit_t *s, stackit_t * const end) {
	stackit_t *lo = NULL;
	
	/* the loop takes 420 bytes of machine code, normal 586, so it's quite small. We can save around 20 bytes for the
	 * loop and the initialization of lo and s->prev and the switch, so ~400 bytes seems reasonable. Multiply with
	 * expected max. number of nodes. If this value is larger than half the L1-cache (due to it's split into code and data),
	 * the optimization should (probably) not be performed. */
	// todo: different cases each with amark0 and amark1 around their respective code in the function
	while (s < end) {
		s->prev = lo;
		lo = s;
		
		switch (s->op) {
			case VAR: lo = _calc_var(s); break;	// x = pop; y = var[x]; push y
			case NUM: case NUMF: break;			// x = pop; push x
			case FUN: lo = _calc_fun(s); break;	// f = pop; x = pop; y = f(x); push y
			case FAC: lo = _calc_fac(s); break;	// ! = pop; x = pop; z = x!; push z
			case EXP: lo = _calc_exp(s); break;	// ^ = pop; y = pop; x = pop; z = x ^ y; push z
			case MUL: lo = _calc_mul(s); break;	// * = pop; y = pop; x = pop; z = x * y; push z
			case DIV: lo = _calc_div(s); break;	// / = pop; y = pop; x = pop; z = x / y; push z
			case ADD: lo = _calc_add(s); break;	// + = pop; y = pop; x = pop; z = x + y; push z
			case SUB: lo = _calc_sub(s); break;	// - = pop; y = pop; x = pop; z = x - y; push z
		}
		s++;
	}
	return;
}

inline static void __fastcall _calc_interprete3(stackit_t *s, stackit_t * const end) {
	stackit_t *lo = NULL, *t;
	
	const int jmptbl[] = {
		&&cvar - &&coff, &&next - &&coff, &&next - &&coff, &&cfun - &&coff,
		&&cfac - &&coff, &&cexp - &&coff, &&cmul - &&coff, &&cdiv - &&coff, &&cadd - &&coff, &&csub - &&coff
	};/*
	const void *jmptbl[] = {
		&&cvar, &&next, &&next, &&cfun,
		&&cfac, &&cexp, &&cmul, &&cdiv, &&cadd, &&csub
	};*/
	
	while (s < end) {
		s->prev = lo;
		lo = s;
		goto *(&&coff + jmptbl[s->op]);
		// goto *jmptbl[s->op];
next:	s++;
	}
	
	amark0(blubb);
	return;
	amark1(blubb);
	
coff:
cvar:
	amark0(var);
	s->data.val = (*s->var)[s->data.idx];
	amark1(var);
	goto next;
	
cfun:
	amark0(fun);
	lo = s->prev;
	lo->data.val = s->data.fun.c(lo->data.val);
	amark1(fun);
	goto next;
	
cfac:
	amark0(fac);
	lo = s->prev;
	lo->data.val = cgamma(lo->data.val, NULL);
	amark1(fac);
	goto next;
	
cexp:
	amark0(exp);
	t = s->prev;
	lo = t->prev;
	lo->data.val = ___cpow(lo->data.val, t->data.val);
	amark1(exp);
	goto next;

#define eval(op, abbr) \
c ## abbr:             \
	amark0(abbr);      \
	t = s->prev;       \
	lo = t->prev;      \
	lo->data.val op ## = t->data.val; \
	amark1(abbr);      \
	goto next;
eval(*, mul)
eval(/, div)
eval(+, add)
eval(-, sub)
#undef eval
}

inline static void __fastcall _calc_interprete4(stackit_t *s, stackit_t * const end) {
	stackit_t *lo = NULL;
	while (s < end) {
		s->prev = lo;
		lo = s->calc(s);
		s++;
	}
}

double complex __fastcall calc(form_t *f, double complex const *var) {
	stackit_t *s, *t, *end;
	
	s = f->stwork;
	end = s + f->node_cnt;
	
	// copy the original data, prev will be set anew below anyway and op never changes, memcpy would be too expensive here
	// todo: move minimize() and the memcpy storg -> stwork into some initialization routine
	t = f->storg;
	do {
		s++->data = t++->data;
	} while (s < end);
	f->var = var;
	
	_calc_interprete4(f->stwork, end);
	
	/*
	unsigned int j;
	for (j=0; j<sizeof(marks)/sizeof(*marks); j++) {
		int i;
		int dm1 = marks[j][1] - marks[j][0];
		if (dm1 == 0) continue;
		printf("code %c:", fsym_chrs[j]);
		for (i=0; i<dm1; i++)
			printf(" %02x", *(marks[j][0] + i) & 0xFF);
		printf("\n");
	}*/
	return f->stwork->data.val;
}

#undef amark0
#undef amark1
#undef _prefix
#undef str

void print_stackit(stackit_t *s, mfuncs_t *mfunctions) {
	double r, c;
	mfuncs_t *mf;
	switch (s->op) {
		case VAR:
			printf("%c", 'x' + s->data.idx);
			break;
		case NUM:
		case NUMF:
			r = creal(s->data.val);
			c = cimag(s->data.val);
			if (r != 0.0 && c != 0.0) {
				printf("(%g+%gi)", r, c);
			} else if (c != 0.0) {
				printf("(%gi)", c);
			} else {
				printf((r >= 0.0 && r < 10.0 && rint(r) == r) ? "%g" : "(%g)", r);
			}
			break;
		
		case FUN:
			for (mf = mfunctions; mf->name != NULL; mf++)
				if (mf->rfun == s->data.fun.r && mf->cfun == s->data.fun.c) {
					printf("{%s}", mf->name);
					break;
				}
			break;
		
		case FAC: case EXP: case MUL:
		case DIV: case ADD: case SUB:
			printf("%c", fsym_chrs[s->op]);
			break;
	}
}

void print_stack(stackit_t *s, int n, mfuncs_t *mfunctions) {
	int i;
	for (i=0; i<n; i++) {
		print_stackit(s, mfunctions);
		s++;
	}
}

void minimize(form_t *f) {
	stackit_t *s;
	int n;
	int sp;
	
	/* for each e as expression of type factor0,factor,term2,term1,expr do
	 *   sort all units of subexpressions with one type lower than e after occurences of variables alphabetically
	 * for each e as expression of type expr,term1,temr2,factor,factor0 do
	 *   collect all units of subexpressions with one type lower than e into a list
	 */
	
	n = f->node_cnt;
	s = f->stwork;
	memcpy(s, f->storg, sizeof(stackit_t) * n);
	
	int cnt = 0;
	double complex b;
	stackit_t *lo = NULL;
	for (sp = 0; sp < n; sp++) {
		s[sp].prev = lo;
		lo = &s[sp];
#define check(t) if (t->prev->op != NUMF && t->prev->op != NUM) break;
		switch (s[sp].op) {
			case VAR:		// don't replace
				break;
				
			case NUM:
			case NUMF:		// nothing to do with constants
				break;
				
			case FUN:		// pre-calculate if it doesn't involve a variable
				check(lo);
				lo = lo->prev;
				lo->data.val = s[sp].data.fun.c(lo->data.val);
				cnt++;
				break;
				
			case FAC:
				check(lo);
				lo = lo->prev;
				lo->data.val = cgamma(lo->data.val, NULL);
				cnt++;
				break;
			case EXP:		// pre-calculate if it doesn't involve a variable
				check(lo);
				check(lo->prev);
				lo = lo->prev;
				b = lo->data.val;
				lo = lo->prev;
				lo->data.val = cpow(lo->data.val, b);
				cnt += 2;
				break;
			
#define eval(op2) do { \
			check(lo); \
			check(lo->prev); \
			lo = lo->prev; \
			b = lo->data.val; \
			lo = lo->prev; \
			lo->data.val op2 ## = b; \
			cnt += 2; \
		} while (0)
			// pre-calculate if it doesn't involve a variable
			case MUL: eval(*); break;
			case DIV: eval(/); break;
			case ADD: eval(+); break;
			case SUB: eval(-); break;
#undef eval
#undef check
		}
	}
	
	static const _c_func funcs[] = {
		_calc_var, _calc_num, _calc_numf, _calc_fun,
		_calc_fac, _calc_exp, _calc_mul, _calc_div, _calc_add, _calc_sub
	};
	
	// copy the stack_it elems back, beginning at the last one, pointed to by lo and following the prev-references
	memset(f->storg, 0, sizeof(stackit_t) * n);
	stackit_t *idx = lo;
	int i = n - cnt;
	// post-init of stack-elements and copy from stwork to storg
	for (i=n-cnt-1; i>=0; i--) {
		idx->var = &f->var;
		idx->calc = funcs[idx->op];
		f->storg[i] = *idx;
		idx = idx->prev;
	}
	f->node_cnt = n - cnt;
	// copy the contiguous stack-elems from storg to stwork
	memcpy(f->stwork, f->storg, f->node_cnt * sizeof(stackit_t));
}

static form_t *cf;
static stackit_t *sit;
static mfuncs_t *mfunctions;
static mvars_t *mvariables;

int form_out_infix = 0;
int form_out_paran = 0;

static void ptree_count(ptree_t *t) {
	if (t != NULL)
		cf->node_cnt++;
}

static void ptree_formfill(ptree_t *t) {
	if (t == NULL) return;
	sit->op = t->sym;
	sit->data.val = 0;
	switch (t->sym) {
		case VAR:
			if (mvariables[t->v.ival].isconstant) {
				sit->op = NUMF;
				sit->data.val = mvariables[t->v.ival].val;
			} else {
				sit->data.idx = t->v.ival;
			}
			break;
		case FUN:
			sit->data.fun.r = mfunctions[t->v.ival].rfun;
			sit->data.fun.c = mfunctions[t->v.ival].cfun;
			break;
		case NUM:
			sit->data.val = (double complex)t->v.ival;
			break;
		case NUMF:
			sit->data.val = (double complex)t->v.fval;
			break;
		default:;
	}
	sit++;
	free(t);
}

void form_prep(const char *str, ptree_t *t, form_t *f, mvars_t *mvars, mfuncs_t *mfuncs, int output) {
	// count number of nodes in t
	// some output so the user knows we've parsed the formula correctly
	cf = f;
	f->node_cnt = 0;
	mfunctions = mfuncs;
	mvariables = mvars;
	
	tnode_vis output_func = (output) ? ptree_output : NULL;
	if (form_out_infix) {
		ptree_output_paran = 1;
		ptree_walk(t, ptree_count, output_func, NULL);
		ptree_output_paran = 0;
	} else {
		ptree_output_paran = form_out_paran;
		ptree_walk(t, ptree_count, NULL, output_func);
		ptree_output_paran = 0;
	}
	
	// allocate the stack space
	f->fstr = strdup(str);
	f->storg = malloc(sizeof(stackit_t) * f->node_cnt);
	f->stwork = malloc(sizeof(stackit_t) * f->node_cnt);
	
	// fill the stack, ptree_formfill sets the constants (NUM) and the pointers to the VARs and FUNCs respectively
	sit = f->storg;
	ptree_walk(t, NULL, NULL, ptree_formfill);
}

void form_free(form_t *f) {
	free(f->fstr);
	free(f->storg);
	free(f->stwork);
}
