
#ifndef _H_EVAL
#define _H_EVAL
#include <stdbool.h>

#define __fastcall __attribute__((fastcall))

#ifndef NEXTERN
extern int form_out_infix;	/* specifies whether the output during form_prep shall occur with in- or postfix notation */
extern int form_out_paran;	/* if above is 0, specifies whether parantheses shall be used for the output */
#endif /* NEXTERN */

typedef double complex (*mcfun_t)(double complex);
typedef double (*mrfun_t)(double);

typedef struct {
	mrfun_t rfun;
	mcfun_t cfun;
	char *name;
	char *desc;
} mfuncs_t;

typedef struct {
	bool isconstant;
	double complex val;
	char *name;
	char *desc;
} mvars_t;

/* This structure contains everything needed to calculate the formula
*  op      same as in the pstree_t-nodes
*  data    holds the data to calculate with (just one of the following fields is used at a time):
*   .val   represents a constant value, op must be NUM or NUMF for it to be valid and interpreted as such
*   .idx   original value as parsed by the formparser out of the function given by the user, op must be VAR or FUN
*   .fun   function-pointers to the real and complex versions of the function as indexed by idx, op must be FUN */
typedef struct stackit_s {
	struct stackit_s *prev;
	fsym op;				// todo: need a way to distinguish between real and complex item?
	union data {
		double complex val;
		int idx;
		struct {
			mcfun_t c;
			mrfun_t r;
		} fun;
	} data;
	struct stackit_s * __fastcall (*calc)(struct stackit_s * const);
	double complex const **var;
} stackit_t;

typedef struct {
	stackit_t *stwork;		// stack to work on, same size as storg, storg's content is copied into it for every evaluation
	stackit_t *storg;		// reference stack for the formula in rev. polish notation
	int node_cnt;			// number of entries in the above stacks (respectively number of nodes in the parse-tree)
	char *fstr;				// the string-representation of the formula as entered by the user
	double complex const *var;
} form_t;


double complex __fastcall calc(form_t *f, double complex const *var);
void print_stackit(stackit_t *s, mfuncs_t *mfunctions);
void print_stack(stackit_t *s, int n, mfuncs_t *mfunctions);
void minimize(form_t *f);
void form_prep(const char *fstr, ptree_t *t, form_t *f, mvars_t *mvars, mfuncs_t *mfuncs, int output);
void form_free(form_t *f);

#endif
