
#ifndef _H_FORMPARSE
#define _H_FORMPARSE

typedef enum {
	VAR, NUM, NUMF, FUN,
	FAC, EXP, MUL, DIV, ADD, SUB
} fsym;

#ifndef NEXTERN
extern const char * const fsym_chrs;		// contains a character symbol for each fsym, "vnNf!^*/+-"
extern int ptree_output_paran;				// set to 1 to enable parantheses around nodes printed by ptree_output
#endif

/*
contents of each "literal" pstree_t-node:
 - var:  tok.value = 0: 'x' | 1: 'y' | 2: 'z'
 - num:  tok.value = numeric value of the number-string
 - numf: tok.value = floating point numeric value of the number-string
 - fun:  tok.value = number of the function as passed to init(), the right child contains the argument for the function
*/

typedef struct ptree_s {
	struct ptree_s *left, *right, *up;
	fsym sym;
	union {
		unsigned int ival;
		float fval;
	} v;
} ptree_t;

/**
* @p s the formula to parse
* @p nvars number of variable-names in @p vars
* @p vars names of variables the caller knows are valid, i.e. 'x', 'y', 'pi', etc.
* @p nfuncs number of function-names in @p funcs
* @p funcs names of functions the callee knows are valid, i.e. 'sin', 'cos', etc.
*/
ptree_t * parse(const char *s, int nvars, char const **vars, int nfuncs, char const **funcs);

typedef void (*tnode_vis)(ptree_t *);
void ptree_walk(ptree_t *t, tnode_vis pre, tnode_vis in, tnode_vis post);
void ptree_output(ptree_t *t);

#endif
